# flake8: noqa
# pylint: disable=broad-exception-raised, raise-missing-from, too-many-arguments, redefined-outer-name
# pylance: disable=reportMissingImports, reportMissingModuleSource, reportGeneralTypeIssues
# type: ignore

import warnings
warnings.filterwarnings("ignore", category=Warning)

import re
from datetime import datetime, timedelta, timezone
import subprocess
import os
import time
import random
import sys
import pkg_resources
import requests
import json
from collections import deque

required_packages = {
    'grist-api': 'latest',
    'colorama': 'latest',
    'requests': '2.31.0',
    'urllib3': '2.0.7',
    'charset-normalizer': '3.3.2'
}

installed_packages = {pkg.key: pkg.version for pkg in pkg_resources.working_set}

for package, version in required_packages.items():
    if package not in installed_packages or (version != 'latest' and installed_packages[package] != version):
        if version == 'latest':
            subprocess.check_call([sys.executable, '-m', 'pip', 'install', package, '--break-system-packages'])
        else:
            subprocess.check_call([sys.executable, '-m', 'pip', 'install', f"{package}=={version}", '--break-system-packages'])

from grist_api import GristDocAPI
import colorama

import logging
import socket

STATE_FILE_PATH = '/root/node/checker_state.json'

def self_update(logger):
    logger.info("Checking for updates..")
    script_path = os.path.abspath(__file__)
    update_url = "https://gitea.vvzvlad.xyz/vvzvlad/ritual/raw/branch/main-22aug/checker.py"
    try:
        response = requests.get(update_url, timeout=10)
        if response.status_code == 200:
            current_content = ""
            with open(script_path, 'r', encoding='utf-8') as f:
                current_content = f.read()
            
            if current_content != response.text:
                with open(script_path, 'w', encoding='utf-8') as f:
                    f.write(response.text)
                logger.info("Script updated successfully, restarting")
                os.execv(sys.executable, ['python3'] + sys.argv)
            else:
                logger.info("Script is up to date")
        else:
            logger.error(f"Failed to download update, status code: {response.status_code}")
    except Exception as e:
        logger.error(f"Update error: {str(e)}")

class GRIST:
    def __init__(self, server, doc_id, api_key, logger):
        self.server = server
        self.doc_id = doc_id
        self.api_key = api_key
        self.logger = logger
        self.grist = GristDocAPI(doc_id, server=server, api_key=api_key)

    def table_name_convert(self, table_name):
        return table_name.replace(" ", "_")

    def to_timestamp(self, dtime: datetime) -> int:
        if dtime.tzinfo is None:
            dtime = dtime.replace(tzinfo=timezone(timedelta(hours=3))) 
        return int(dtime.timestamp())

    def insert_row(self, data, table):
        data = {key.replace(" ", "_"): value for key, value in data.items()}
        row_id = self.grist.add_records(self.table_name_convert(table), [data])
        return row_id

    def update_column(self, row_id, column_name, value, table):
        if isinstance(value, datetime):
            value = self.to_timestamp(value)
        column_name = column_name.replace(" ", "_")
        self.grist.update_records(self.table_name_convert(table), [{ "id": row_id, column_name: value }])

    def delete_row(self, row_id, table):
        self.grist.delete_records(self.table_name_convert(table), [row_id])

    def update(self, row_id, updates, table):
        for column_name, value in updates.items():
            if isinstance(value, datetime):
                updates[column_name] = self.to_timestamp(value)
        updates = {column_name.replace(" ", "_"): value for column_name, value in updates.items()}
        self.grist.update_records(self.table_name_convert(table), [{"id": row_id, **updates}])

    def fetch_table(self, table):
        return self.grist.fetch_table(self.table_name_convert(table))

    def find_record(self, id=None, state=None, name=None, table=None):
        if table is None:
            raise ValueError("Table is not specified")
        table_data = self.grist.fetch_table(self.table_name_convert(table))
        if id is not None:
            record = [row for row in table_data if row.id == id]
            return record
        if state is not None and name is not None:
            record = [row for row in table_data if row.State == state and row.name == name]
            return record
        if state is not None:
            record = [row for row in table_data if row.State == state]
            return record
        if name is not None:
            record = [row for row in table_data if row.Name == name]
            return record

    def find_settings(self, key, table):
        table = self.fetch_table(self.table_name_convert(table))
        for record in table:
            if record.Setting == key:
                if record.Value is None or record.Value == "":
                    raise ValueError(f"Setting {key} blank")
                return record.Value
        raise ValueError(f"Setting {key} not found")

def read_state(logger):
    """Reads the last known state (status and sync count) from the state file."""
    default_state = {"last_status": "Idle", "sync_count": 0}
    if not os.path.exists(STATE_FILE_PATH):
        logger.info(f"State file {STATE_FILE_PATH} not found, starting with default state.")
        return default_state
    try:
        with open(STATE_FILE_PATH, 'r', encoding='utf-8') as f:
            state = json.load(f)
            # Validate state structure
            if "last_status" not in state or "sync_count" not in state:
                logger.warning(f"Invalid state file format in {STATE_FILE_PATH}, using default state.")
                return default_state
            # Ensure sync_count is an integer
            state["sync_count"] = int(state["sync_count"])
            return state
    except (json.JSONDecodeError, ValueError, IOError) as e:
        logger.error(f"Error reading or parsing state file {STATE_FILE_PATH}: {e}. Using default state.")
        return default_state

def write_state(state, logger):
    """Writes the current state to the state file."""
    try:
        with open(STATE_FILE_PATH, 'w', encoding='utf-8') as f:
            json.dump(state, f, indent=4)
    except IOError as e:
        logger.error(f"Error writing state file {STATE_FILE_PATH}: {e}")

def clean_ansi(text):
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    return ansi_escape.sub('', text)

def format_number(number_str):
    try:
        number = int(number_str)
        if number >= 1000:
            return f"{number//1000}k"
        return str(number)
    except (ValueError, TypeError):
        return "NaN" # Or some other indicator of invalid input

def check_logs(logger):
    state = read_state(logger)
    original_sync_count = state["sync_count"]
    previous_status = state["last_status"]
    
    try:
        logs = subprocess.run(['docker', 'logs', '--since', '10m', 'infernet-node'], capture_output=True, text=True, check=True)
        log_content = clean_ansi(logs.stdout)

        last_checking_info = None
        last_ignored_id = None
        last_head_sub_id = None

        # Regex patterns
        checking_pattern = re.compile(r'Checking subscriptions.*last_sub_id=(\d+).*head_sub_id=(\d+).*num_subs_to_sync=(\d+)')
        ignored_pattern = re.compile(r'Ignored subscription creation.*id=(\d+)')
        head_sub_pattern = re.compile(r'head sub id is:\s*(\d+)')

        # Use deque to efficiently get the last few relevant lines if needed,
        # but processing all lines and keeping the last match is simpler here.
        for line in log_content.splitlines():
            match = checking_pattern.search(line)
            if match:
                last_checking_info = {
                    "last_sub_id": match.group(1),
                    "head_sub_id": match.group(2),
                    "num_subs_to_sync": int(match.group(3))
                }
                continue # Prioritize checking_info

            match = ignored_pattern.search(line)
            if match:
                last_ignored_id = match.group(1)
                continue

            match = head_sub_pattern.search(line)
            if match:
                 last_head_sub_id = match.group(1)
                 # No continue here, allows checking_info from same timeframe to override

        current_status_type = "Idle"
        status_message = "Idle"
        
        if last_checking_info:
            formatted_id = format_number(last_checking_info["last_sub_id"])
            if last_checking_info["num_subs_to_sync"] > 0:
                current_status_type = "Sync"
                status_message = f"Sync: {formatted_id} ({state['sync_count']})"
                logger.info(f"Node is syncing. Last sub ID: {last_checking_info['last_sub_id']}, Num subs to sync: {last_checking_info['num_subs_to_sync']}")
            else:
                current_status_type = "OK"
                # Increment count only on transition from Sync to OK based on the *last* status type
                if previous_status == "Sync":
                    state["sync_count"] += 1
                    logger.info(f"Sync completed. Sync count incremented to {state['sync_count']}.")
                status_message = f"OK: {formatted_id} ({state['sync_count']})"
                logger.info(f"Node is OK. Last sub ID: {last_checking_info['last_sub_id']}")
        
        elif last_ignored_id:
            # Fallback to "Ignored" logs if "Checking" is missing
            formatted_id = format_number(last_ignored_id)
            current_status_type = "Sync" # Assume sync if we only see ignored creations recently
            status_message = f"Sync: {formatted_id} ({state['sync_count']})"
            logger.info(f"Node possibly syncing (based on ignored logs). Last ignored ID: {last_ignored_id}")

        elif last_head_sub_id:
            # Fallback to "head sub id" if others are missing
            formatted_id = format_number(last_head_sub_id)
            current_status_type = "OK" # Assume OK if this is the latest relevant info
            status_message = f"OK: {formatted_id} ({state['sync_count']})"
            logger.info(f"Node status based on head sub id. Head sub ID: {last_head_sub_id}")
            
        else:
            logger.info("No relevant subscription log entries found in the last 10 minutes. Status: Idle.")
            status_message = "Idle"
            current_status_type = "Idle"

        # Update state and write back if changed
        state["last_status"] = current_status_type
        if state["sync_count"] != original_sync_count or state["last_status"] != previous_status:
            write_state(state, logger)

        return {"status": status_message}

    except subprocess.CalledProcessError as e:
        logger.error(f"Error running docker logs command: {e.stderr or e.stdout or e}")
        # Don't update state file on command error, keep previous state
        return {"status": f"Error: Docker logs failed ({e.returncode})"}
    except Exception as e:
        logger.error(f"Unexpected error processing logs: {e}", exc_info=True)
        # Don't update state file on unexpected error
        return {"status": f"Error: Log processing failed"}

if __name__ == "__main__":
    colorama.init(autoreset=True)
    logger = logging.getLogger("Checker")
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    logger.info("Checker started")
    self_update(logger)
    #random_sleep = random.randint(1, 60)
    #logger.info(f"Sleeping for {random_sleep} seconds")
    #time.sleep(random_sleep)

    grist_data = {}
    with open('/root/node/grist.json', 'r', encoding='utf-8') as f:
        grist_data = json.loads(f.read())

    GRIST_ROW_NAME = socket.gethostname()
    NODES_TABLE = "Nodes"
    grist = GRIST(grist_data.get('grist_server'), grist_data.get('grist_doc_id'), grist_data.get('grist_api_key'), logger)
    current_vm = grist.find_record(name=GRIST_ROW_NAME, table=NODES_TABLE)[0]
    def grist_callback(msg): grist.update(current_vm.id, msg, NODES_TABLE)

    for attempt in range(3):
        try:        
            result = check_logs(logger)
            grist_callback({ "Health": result["status"] })
            logger.info(f"Status: {result['status']}")
            break
        except Exception as e:  
            logger.error(f"Error on attempt {attempt+1}/3: {e}")
            if attempt == 2:
                grist_callback({ "Health": f"Error: {e}" })
            if attempt < 2:
                time.sleep(5)